from .constraint import Constraint
from .exclusive_nodes import ExclusiveNodes
from .expression import ExpressionConstraint
from .max_children import MaxChildren
from .max_parents import MaxParents
from .pin import Pin

__all__ = [
    "Constraint",
    "ExclusiveNodes",
    "ExpressionConstraint",
    "MaxChildren",
    "MaxParents",
    "Pin",
]



from .appear import Appear
from .cost import Cost
from .disappear import Disappear
from .edge_distance import EdgeDistance
from .edge_selection import EdgeSelection
from .features import Features
from .merge import Merge
from .node_selection import NodeSelection
from .split import Split
from .weight import Weight
from .weights import Weights

__all__ = [
    "Appear",
    "Cost",
    "Disappear",
    "EdgeDistance",
    "EdgeSelection",
    "Features",
    "Merge",
    "NodeSelection",
    "Split",
    "Weight",
    "Weights",
]



from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING, Callable, Mapping, TypeVar, cast

import ilpy
import numpy as np

from .constraints.constraint import Constraint
from .costs import Features, Weight, Weights
from .ssvm import fit_weights
from .track_graph import TrackGraph

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from motile.costs import Cost
    from motile.variables import Variable

    V = TypeVar("V", bound=Variable)


class Solver:
    """Create a solver for a given track graph.

    Args:
        track_graph:
            The :class:`~motile.TrackGraph` of objects to track over time.
    """

    def __init__(self, track_graph: TrackGraph) -> None:
        if not isinstance(track_graph, TrackGraph):
            import networkx as nx

            if isinstance(track_graph, nx.Graph):
                warnings.warn(
                    "Coercing networkx graph to TrackGraph with frame_attribute='t'. "
                    "To silence this warning, please pass a motile.TrackGraph instance."
                )
                track_graph = TrackGraph(track_graph)
            else:  # pragma: no cover
                raise ValueError(
                    f"Expected a TrackGraph or networkx.Graph, got {type(track_graph)}"
                )

        self.graph = track_graph
        self.variables: dict[type[Variable], Variable] = {}
        self.variable_types: dict[int, ilpy.VariableType] = {}

        self.weights = Weights()
        self.weights.register_modify_callback(self._on_weights_modified)
        self._weights_changed = True
        self.features = Features()

        self.ilp_solver: ilpy.Solver | None = None
        self.objective: ilpy.Objective | None = None
        self.constraints = ilpy.Constraints()

        self.num_variables: int = 0
        self._costs = np.zeros((0,), dtype=np.float32)
        self._cost_instances: dict[str, Cost] = {}
        self.solution: ilpy.Solution | None = None

    def add_cost(self, cost: Cost, name: str | None = None) -> None:
        """Add linear cost to the value of variables in this solver.

        Args:
            cost:
                The cost to add.  An instance of :class:`~motile.costs.Cost`.

            name:
                An optional name of the , used to refer to weights of
                cost in an unambiguous manner. Defaults to the name of the
                cost class, if not given.
        """
        # default name of the cost is the class name
        if name is None:
            name = type(cost).__name__

        if name in self._cost_instances:
            raise RuntimeError(
                f"A cost instance with name '{name}' was already registered. "
                "Consider passing a different name with the 'name=' argument "
                "to Solver.add_cost"
            )

        logger.info("Adding %s cost...", name)

        self._cost_instances[name] = cost

        # fish out all weights used in this cost object
        for var_name, var in cost.__dict__.items():
            if not isinstance(var, Weight):
                continue
            weight_index = self.weights.add_weight(var, (name, var_name))
            self.features.register_feature(weight_index)

        cost.apply(self)

    def add_constraint(self, constraint: Constraint) -> None:
        """Add linear constraints to the solver.

        Args:
            constraint:
                The :class:`~motile.constraints.Constraint` to add.
        """
        logger.info("Adding %s constraint...", type(constraint).__name__)

        for constraint in constraint.instantiate(self):
            self.constraints.add(constraint)

    def solve(
        self,
        timeout: float = 0.0,
        num_threads: int = 1,
        verbose: bool = False,
        backend: ilpy.Preference = ilpy.Preference.Any,
        on_event: Callable[[Mapping], None] | None = None,
    ) -> ilpy.Solution:
        """Solve the global optimization problem.

        Args:
            timeout:
                The timeout for the ILP solver, in seconds. Default (0.0) is no
                timeout. If the solver times out, the best solution encountered
                so far is returned (if any has been found at all).
            num_threads:
                The number of threads the ILP solver uses.
            verbose:
                If true, print more information from ILP solver. Defaults to False.
            backend:
                The ILP solver backend to use. Defaults to Any.
            on_event:
                A callback function that will be called when the solver emits an event.
                Should accept an event data dict. (see `ilpy.EventData` for typing info
                which may be imported inside of a TYPE_CHECKING block.)
                Defaults to None.

        Returns:
            :class:`ilpy.Solution`, a vector of variable values. Use
            :func:`get_variables` to find the indices of variables in this
            vector.
        """
        self.objective = ilpy.Objective(self.num_variables)
        for i, c in enumerate(self.costs):
            logger.debug("Setting cost of var %d to %.3f", i, c)
            self.objective.set_coefficient(i, c)

        # TODO: support other variable types
        self.ilp_solver = ilpy.Solver(
            self.num_variables,
            ilpy.VariableType.Binary,
            variable_types=self.variable_types,
            preference=backend,
        )

        self.ilp_solver.set_objective(self.objective)
        self.ilp_solver.set_constraints(self.constraints)

        self.ilp_solver.set_num_threads(num_threads)
        if timeout > 0:
            self.ilp_solver.set_timeout(timeout)

        self.ilp_solver.set_verbose(verbose)
        self.ilp_solver.set_event_callback(on_event)

        self.solution = self.ilp_solver.solve()
        if message := self.solution.get_status():
            logger.info("ILP solver returned with: %s", message)

        return self.solution

    def get_variables(self, cls: type[V]) -> V:
        """Get variables by their class name.

        If the solver does not yet contain those variables, they will be
        created.

        Args:
            cls (type of :class:`motile.variables.Variable`):
                A subclass of :class:`motile.variables.Variable`.

        Returns:
            A singleton instance of :class:`~motile.variables.Variable` (of whatever
            type was passed in as ``cls``), mimicking a dictionary that can be used to
            look up variable indices by their keys. See
            :class:`~motile.variables.Variable` for details.
        """
        if cls not in self.variables:
            self._add_variables(cls)
        return cast("V", self.variables[cls])

    def add_variable_cost(
        self, index: int | ilpy.Variable, value: float, weight: Weight
    ) -> None:
        """Add cost for an individual variable.

        To be used within implementations of :class:`motile.costs.Cost`.
        """
        variable_index = index
        feature_index = self.weights.index_of(weight)
        self.features.add_feature(variable_index, feature_index, value)

    def fit_weights(
        self,
        gt_attribute: str,
        regularizer_weight: float = 0.1,
        max_iterations: int = 1000,
        eps: float = 1e-6,
    ) -> None:
        """Fit weights of ILP costs to ground truth with structured SVM.

        Updates the weights in the solver object to the found solution.

        See https://github.com/funkelab/structsvm for details.

        Args:
            gt_attribute:
                Node/edge attribute that marks the ground truth for fitting.
                `gt_attribute` is expected to be set to:

                - ``1`` for objects labaled as ground truth.
                - ``0`` for objects explicitly labeled as not part of the ground truth.
                - ``None`` or not set for unlabeled objects.

            regularizer_weight:
                The weight of the quadratic regularizer.

            max_iterations:
                Maximum number of gradient steps in the structured SVM.

            eps:
                Convergence threshold.
        """
        optimal_weights = fit_weights(
            self, gt_attribute, regularizer_weight, max_iterations, eps
        )
        self.weights.from_ndarray(optimal_weights)

    @property
    def costs(self) -> np.ndarray:
        """Returns the costs as a :class:`numpy.ndarray`."""
        if self._weights_changed:
            self._compute_costs()
            self._weights_changed = False

        return self._costs

    def _add_variables(self, cls: type[V]) -> None:
        logger.info("Adding %s variables...", cls.__name__)

        keys = cls.instantiate(self)
        indices = self._allocate_variables(len(keys))
        variables = cls(self, dict(zip(keys, indices)))
        self.variables[cls] = variables

        for index in indices:
            self.variable_types[index] = cls.variable_type

        for constraint in cls.instantiate_constraints(self):
            self.constraints.add(constraint)

        self.features.resize(num_variables=self.num_variables)

    def _compute_costs(self) -> None:
        logger.info("Computing costs...")

        weights = self.weights.to_ndarray()
        features = self.features.to_ndarray()
        self._costs = np.dot(features, weights)

    def _allocate_variables(self, num_variables: int) -> range:
        indices = range(self.num_variables, self.num_variables + num_variables)

        self.num_variables += num_variables
        self.features.resize(num_variables=self.num_variables)

        return indices

    def _on_weights_modified(self, old_value: float | None, new_value: float) -> None:
        if old_value != new_value:
            if not self._weights_changed:
                logger.info("Weights have changed")

            self._weights_changed = True

    def get_selected_subgraph(
        self, solution: ilpy.Solution | None = None
    ) -> TrackGraph:
        """Return TrackGraph with only the selected nodes/edges from the solution.

        Args:
            solution:
                The solution to use. If not provided, the last solution is used.

        Returns:
            A new TrackGraph with only the selected nodes and edges.

        Raises:
            RuntimeError: If no solution is provided and the solver has not been solved
            yet.
        """
        from motile.variables import EdgeSelected, NodeSelected

        if solution is None:
            solution = self.solution

        # TODO:
        # in theory this could be made more efficient by using a nx.DiGraph view
        # but TrackGraph itself doesn't provide views (and isn't a subclass)
        if not solution:
            raise RuntimeError(
                "No solution available. Run solve() first or manually pass a solution."
            )

        node_selected = self.get_variables(NodeSelected)
        edge_selected = self.get_variables(EdgeSelected)
        selected_graph = TrackGraph(frame_attribute=self.graph.frame_attribute)

        for node_id, node in self.graph.nodes.items():
            if solution[node_selected[node_id]] > 0.5:
                selected_graph.add_node(node_id, node)

        for edge_id, edge in self.graph.edges.items():
            if solution[edge_selected[edge_id]] > 0.5:
                selected_graph.add_edge(edge_id, edge)

        return selected_graph



from __future__ import annotations

from typing import Collection, Hashable, Iterable

import ilpy
import pytest
from motile import Solver, TrackGraph
from motile.variables import Variable


@pytest.mark.parametrize("VarCls", Variable.__subclasses__())
def test_variable_subclass_protocols(
    arlo_graph: TrackGraph, VarCls: type[Variable]
) -> None:
    """Test that all Variable subclasses properly implement the Variable protocol."""
    solver = Solver(arlo_graph)

    keys = VarCls.instantiate(solver)
    assert isinstance(keys, Collection)
    assert all(isinstance(k, Hashable) for k in keys)

    constraints = VarCls.instantiate_constraints(solver)
    assert isinstance(constraints, Iterable)
    assert all(isinstance(c, (ilpy.Expression, ilpy.Constraint)) for c in constraints)



import networkx as nx
import pytest
from motile import TrackGraph


@pytest.fixture
def arlo_graph_nx() -> nx.DiGraph:
    """Create the "Arlo graph", a simple toy graph for testing.

       x
       |
    200|           6
       |         /
    150|   1---3---5
       |     x   x
    100|   0---2---4
        ------------------------------------ t
           0   1   2
    """
    cells = [
        {"id": 0, "t": 0, "x": 101, "score": 1.0},
        {"id": 1, "t": 0, "x": 150, "score": 1.0},
        {"id": 2, "t": 1, "x": 100, "score": 1.0},
        {"id": 3, "t": 1, "x": 151, "score": 1.0},
        {"id": 4, "t": 2, "x": 102, "score": 1.0},
        {"id": 5, "t": 2, "x": 149, "score": 1.0},
        {"id": 6, "t": 2, "x": 200, "score": 1.0},
    ]

    edges = [
        {"source": 0, "target": 2, "prediction_distance": 1.0},
        {"source": 1, "target": 3, "prediction_distance": 1.0},
        {"source": 0, "target": 3, "prediction_distance": 50.0},
        {"source": 1, "target": 2, "prediction_distance": 50.0},
        {"source": 2, "target": 4, "prediction_distance": 2.0},
        {"source": 3, "target": 5, "prediction_distance": 2.0},
        {"source": 2, "target": 5, "prediction_distance": 49.0},
        {"source": 3, "target": 4, "prediction_distance": 49.0},
        {"source": 3, "target": 6, "prediction_distance": 3.0},
    ]

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    nx_graph.add_edges_from([(edge["source"], edge["target"], edge) for edge in edges])
    return nx_graph


@pytest.fixture
def arlo_graph(arlo_graph_nx) -> TrackGraph:
    """Return the "Arlo graph" as a :class:`motile.TrackGraph` instance."""
    return TrackGraph(arlo_graph_nx)


@pytest.fixture
def toy_graph_nx() -> nx.DiGraph:
    """Return variation of the "Arlo graph".

    Relative to arlo_graph, this graph has:
    - one simple edge modified.
    - normalized node and edge scores.
    - sparse ground truth annotations.

       x
       |
       |       --- 6
       |     /   /
       |   1---3---5
       |     /   x
       |   0---2---4
        ------------------------------------ t
           0   1   2
    """
    cells = [
        {"id": 0, "t": 0, "x": 1, "score": 0.8, "gt": 1},
        {"id": 1, "t": 0, "x": 25, "score": 0.1},
        {"id": 2, "t": 1, "x": 0, "score": 0.3, "gt": 1},
        {"id": 3, "t": 1, "x": 26, "score": 0.4},
        {"id": 4, "t": 2, "x": 2, "score": 0.6, "gt": 1},
        {"id": 5, "t": 2, "x": 24, "score": 0.3, "gt": 0},
        {"id": 6, "t": 2, "x": 35, "score": 0.7},
    ]
    edges = [
        {"source": 0, "target": 2, "score": 0.9, "gt": 1},
        {"source": 1, "target": 3, "score": 0.9},
        {"source": 0, "target": 3, "score": 0.5},
        {"source": 1, "target": 6, "score": 0.5},
        {"source": 2, "target": 4, "score": 0.7, "gt": 1},
        {"source": 3, "target": 5, "score": 0.7},
        {"source": 2, "target": 5, "score": 0.3, "gt": 0},
        {"source": 3, "target": 4, "score": 0.3},
        {"source": 3, "target": 6, "score": 0.8},
    ]
    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    nx_graph.add_edges_from([(edge["source"], edge["target"], edge) for edge in edges])
    return nx_graph


@pytest.fixture
def toy_graph(toy_graph_nx) -> TrackGraph:
    """Return the `toy_graph_nx` as a :class:`motile.TrackGraph` instance."""
    return TrackGraph(toy_graph_nx)


@pytest.fixture
def toy_hypergraph_nx() -> nx.DiGraph:
    """Return variation of `toy_graph` with an edge modified and one hyperedge added.

    Visually:

        x
        |
        |       --- 6
        |     /   /
        |   1---3---5
        |     /   x
        |   0---2---4  Hyperedge: (0,(2,3))
        ------------------------------------ t
            0   1   2
    """
    cells = [
        {"id": 0, "t": 0, "x": 1, "score": 0.8, "gt": 1},
        {"id": 1, "t": 0, "x": 25, "score": 0.1},
        {"id": 2, "t": 1, "x": 0, "score": 0.3, "gt": 1},
        {"id": 3, "t": 1, "x": 26, "score": 0.4},
        {"id": 4, "t": 2, "x": 2, "score": 0.6, "gt": 1},
        {"id": 5, "t": 2, "x": 24, "score": 0.3, "gt": 0},
        {"id": 6, "t": 2, "x": 35, "score": 0.7, "gt": None},
    ]

    edges = [
        {"source": 0, "target": 2, "score": 0.9, "gt": 1},
        {"source": 1, "target": 3, "score": 0.9},
        {"source": 0, "target": 3, "score": 0.5},
        {"source": 1, "target": 6, "score": 0.2},
        {"source": 2, "target": 4, "score": 0.7, "gt": 1},
        {"source": 3, "target": 5, "score": 0.7},
        {"source": 2, "target": 5, "score": 0.3, "gt": 0},
        {"source": 3, "target": 4, "score": 0.3},
        {"source": 3, "target": 6, "score": 0.8, "gt": None},
    ]

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])  # type: ignore
    nx_graph.add_edges_from(
        [(edge["source"], edge["target"], edge) for edge in edges]  # type: ignore
    )

    # this is how to add a TrackGraph hyperedge into a nx_graph:
    nx_graph.add_node(
        10, division_score=0.1
    )  # is a hypernode, because it has no frame attribute
    nx_graph.add_edge(0, 10)
    nx_graph.add_edge(10, 2)
    nx_graph.add_edge(10, 3)
    return nx_graph


@pytest.fixture
def toy_hypergraph(toy_hypergraph_nx) -> TrackGraph:
    """Return the `toy_hypergraph_nx` as a :class:`motile.TrackGraph` instance."""
    return TrackGraph(toy_hypergraph_nx)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import ilpy


class Features:
    """Simple container for features with resizeable dimensions.

    A :class:`~motile.Solver` has a :class:`Features` instance.
    """

    def __init__(self) -> None:
        self._values = np.zeros((0, 0), dtype=np.float32)

    def resize(
        self, num_variables: int | None = None, num_features: int | None = None
    ) -> None:
        """Resize the feature matrix.

        Args:
            num_variables:
                The number of variables to resize to. If None, the number of
                variables is not changed.
            num_features:
                The number of features to resize to. If None, the number of
                features is not changed.
        """
        if num_variables is None:
            num_variables = self._values.shape[0]
        if num_features is None:
            num_features = self._values.shape[1]

        # Resize features first for efficiency
        if num_features > self._values.shape[1]:
            self._increase_features(num_features)

        if num_variables > self._values.shape[0]:
            self._increase_variables(num_variables)

    def _increase_variables(self, num_variables: int) -> None:
        # Increasing size without copying works only in dim 0
        self._values.resize((num_variables, self._values.shape[1]), refcheck=False)

    def _increase_features(self, num_features: int) -> None:
        # Need to copy the array when increasing size of dim 1
        shape = (self._values.shape[0], num_features - self._values.shape[1])
        new_features = np.zeros(shape, dtype=self._values.dtype)
        self._values = np.hstack((self._values, new_features))

    def register_feature(self, feature_index: int) -> None:
        num_variables, num_features = self._values.shape
        if feature_index >= num_features:
            self.resize(
                num_variables,
                max(feature_index + 1, num_features),
            )

    def add_feature(
        self, variable_index: int | ilpy.Variable, feature_index: int, value: float
    ) -> None:
        """Add a value to a feature.

        Args:
            variable_index:
                The index of the variable to add the value to.
            feature_index:
                The index of the feature to add the value to.
            value:
                The value to add.
        """
        num_variables, num_features = self._values.shape

        variable_index = int(variable_index)
        if variable_index >= num_variables:
            self.resize(
                max(variable_index + 1, num_variables),
                num_features,
            )

        self._values[variable_index, feature_index] += value

    def to_ndarray(self) -> np.ndarray:
        """Export the feature matrix as a numpy array.

        Note: you can also use ``np.asarray(features)``.
        """
        # _values is already an ndarray, but this might change in the future
        # Note: consider implementing
        return self._values

    def __array__(self) -> np.ndarray:
        return self.to_ndarray()

    def __repr__(self) -> str:
        r = f"array of shape={self._values.shape}"
        if self._values.size > 0:
            r += f", values =\n{self._values}"
        else:
            r += " [empty]"
        return r



from __future__ import annotations

from .expression import ExpressionConstraint


class Pin(ExpressionConstraint):
    """Enforces the selection of nodes/edges based on truthiness of a given attribute.

    Every node or edge that has the given attribute will be selected if the
    attribute value is ``True`` (and not selected if the attribute value is
    ``False``). The solver will only determine the selection of nodes and edges
    that do not have this attribute.

    This constraint is useful to complete partial tracking solutions, e.g.,
    when users have already manually selected (or unselected) certain nodes or
    edges.

    Args:
        attribute:
            The name of the node/edge attribute to use.
    """

    def __init__(self, attribute: str) -> None:
        super().__init__(f"{attribute} == True", eval_nodes=True, eval_edges=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Collection, Iterable

from .edge_selected import EdgeSelected
from .node_selected import NodeSelected
from .variable import Variable

if TYPE_CHECKING:
    import ilpy

    from motile._types import Node
    from motile.solver import Solver


class NodeDisappear(Variable["Node"]):
    r"""Binary variable to indicate whether a node disappears.

    This variable indicates whether the node is the end of a track (i.e., the node is
    selected and has no selected outgoing edges).

    This variable is coupled to the node and edge selection variables through
    the following linear constraints:

    .. math::
        |\\text{out_edges}(v)|\cdot x_v - &\sum_{e \in \\text{out_edges}(v)} x_e
        - d_v &\leq&\;\; |\\text{out_edges}(v)| - 1

        |\\text{out_edges}(v)|\cdot x_v - &\sum_{e \in \\text{out_edges}(v)} x_e
        - d_v\cdot |\\text{out_edges}(v)| &\geq&\;\; 0

    where :math:`x_v` and :math:`x_e` are selection indicators for node
    :math:`v` and edge :math:`e`, and :math:`d_v` is the disappear indicator for
    node :math:`v`.
    """

    @staticmethod
    def instantiate(solver: Solver) -> Collection[Node]:
        return solver.graph.nodes

    @staticmethod
    def instantiate_constraints(solver: Solver) -> Iterable[ilpy.Expression]:
        node_indicators = solver.get_variables(NodeSelected)
        edge_indicators = solver.get_variables(EdgeSelected)
        disappear_indicators = solver.get_variables(NodeDisappear)

        for node in solver.graph.nodes:
            next_edges = solver.graph.next_edges[node]
            selected = node_indicators[node]
            disappear = disappear_indicators[node]

            if not next_edges:
                # special case: no outgoing edges, disappear indicator is equal to
                # selection indicator
                yield selected == disappear
                continue

            # Ensure that the following holds:
            #
            # disappear = 1 <=> sum(next_selected) = 0 and selected
            # disappear = 0 <=> sum(next_selected) > 0 or not selected
            #
            # Two linear constraints are needed for that:
            #
            # let s = num_next * selected - sum(next_selected)
            # (1) s - disappear <= num_next - 1
            # (2) s - disappear * num_next >= 0

            num_next = len(next_edges)
            s = num_next * selected - sum(edge_indicators[e] for e in next_edges)

            yield s - disappear <= num_next - 1
            yield s - disappear >= 0



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from ..variables import NodeSelected
from .constraint import Constraint

if TYPE_CHECKING:
    import ilpy

    from motile.solver import Solver


class ExclusiveNodes(Constraint):
    r"""Ensures that for each given set of nodes at most one node is selected.

    Adds the following linear constraint for each given exclusive set :math:`S`:

    .. math::

      \sum_{n \in S} x_n \leq 1

    Args:
        exclusive_sets:
            A list of sets of nodes that are mutually exclusive.
    """

    def __init__(self, exclusive_sets: Iterable[Iterable]) -> None:
        self.exclusive_sets = exclusive_sets

    def instantiate(self, solver: Solver) -> Iterable[ilpy.Expression]:
        node_indicators = solver.get_variables(NodeSelected)
        for exclusive_set in self.exclusive_sets:
            yield sum([node_indicators[n] for n in exclusive_set]) <= 1



from __future__ import annotations

from typing import TYPE_CHECKING, Collection, Iterable

from .node_selected import NodeSelected
from .variable import Variable

if TYPE_CHECKING:
    import ilpy

    from motile._types import GenericEdge
    from motile.solver import Solver


class EdgeSelected(Variable["GenericEdge"]):
    """Binary variable indicates whether an edge is part of the solution or not."""

    @staticmethod
    def instantiate(solver: Solver) -> Collection[GenericEdge]:
        return solver.graph.edges

    @staticmethod
    def instantiate_constraints(solver: Solver) -> Iterable[ilpy.Expression]:
        node_indicators = solver.get_variables(NodeSelected)
        edge_indicators = solver.get_variables(EdgeSelected)

        for edge in solver.graph.edges:
            nodes = list(solver.graph.nodes_of(edge))
            x_e = edge_indicators[edge]
            yield len(nodes) * x_e - sum(node_indicators[n] for n in nodes) <= 0



import logging

import motile
import networkx
import numpy as np
from motile.constraints import MaxChildren, MaxParents
from motile.costs import Appear, EdgeSelection, NodeSelection
from motile.variables import EdgeSelected, NodeSelected

logger = logging.getLogger(__name__)


def create_ssvm_noise_trackgraph() -> motile.TrackGraph:
    cells = [
        {"id": 0, "t": 0, "x": 1, "score": 0.8, "gt": 1, "noise": 0.5},
        {"id": 1, "t": 0, "x": 25, "score": 0.9, "gt": 1, "noise": -0.5},
        {"id": 2, "t": 1, "x": 0, "score": 0.9, "gt": 1, "noise": 0.5},
        {"id": 3, "t": 1, "x": 26, "score": 0.8, "gt": 1, "noise": -0.5},
        {"id": 4, "t": 2, "x": 2, "score": 0.9, "gt": 1, "noise": 0.5},
        {"id": 5, "t": 2, "x": 24, "score": 0.1, "gt": 0, "noise": -0.5},
        {"id": 6, "t": 2, "x": 35, "score": 0.7, "gt": 1, "noise": -0.5},
    ]

    edges = [
        {"source": 0, "target": 2, "score": 0.9, "gt": 1, "noise": 0.5},
        {"source": 1, "target": 3, "score": 0.9, "gt": 1, "noise": -0.5},
        {"source": 0, "target": 3, "score": 0.2, "gt": 0, "noise": 0.5},
        {"source": 1, "target": 2, "score": 0.2, "gt": 0, "noise": -0.5},
        {"source": 2, "target": 4, "score": 0.9, "gt": 1, "noise": 0.5},
        {"source": 3, "target": 5, "score": 0.1, "gt": 0, "noise": -0.5},
        {"source": 2, "target": 5, "score": 0.2, "gt": 0, "noise": 0.5},
        {"source": 3, "target": 4, "score": 0.2, "gt": 0, "noise": -0.5},
        {"source": 3, "target": 6, "score": 0.8, "gt": 1, "noise": -0.5},
    ]
    graph = networkx.DiGraph()
    graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    graph.add_edges_from([(edge["source"], edge["target"], edge) for edge in edges])

    return motile.TrackGraph(graph)


def create_toy_solver(graph):
    solver = motile.Solver(graph)

    solver.add_constraint(MaxParents(1))
    solver.add_constraint(MaxChildren(1))

    solver.add_cost(NodeSelection(weight=1.0, attribute="score", constant=-10.0))
    solver.add_cost(EdgeSelection(weight=10.0, attribute="score"))
    solver.add_cost(Appear(constant=10.0))

    logger.debug("====== Initial Weights ======")
    logger.debug(solver.weights)

    return solver


def test_structsvm_common_toy_example(toy_graph):
    graph = toy_graph

    solver = create_toy_solver(graph)

    logger.debug("====== Initial Solution ======")
    solver.solve()
    logger.debug(solver.get_variables(EdgeSelected))

    # Structured Learning
    solver.fit_weights(
        gt_attribute="gt", regularizer_weight=0.03, max_iterations=50, eps=1e-6
    )

    logger.debug("====== Learnt Weights ======")
    logger.debug(solver.weights)
    logger.debug("====== Final Solution ======")
    optimal_weights = solver.weights

    np.testing.assert_allclose(
        optimal_weights[("NodeSelection", "weight")], -3.27, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("NodeSelection", "constant")], 1.78, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("EdgeSelection", "weight")], -3.23, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("EdgeSelection", "constant")], 1.06, atol=0.01
    )
    np.testing.assert_allclose(optimal_weights[("Appear", "constant")], 0.20, atol=0.01)

    solver = create_toy_solver(graph)
    solver.weights.from_ndarray(optimal_weights.to_ndarray())

    logger.debug("====== Learnt Weights in new solver ======")
    logger.debug(solver.weights)
    solution = solver.solve()
    logger.debug(solver.get_variables(EdgeSelected))

    edge_indicators = solver.get_variables(EdgeSelected)
    selected_edges = [
        edge for edge, index in edge_indicators.items() if solution[index] > 0.5
    ]
    for (u, v), attrs in graph.edges.items():
        gt = attrs.get("gt", None)
        if gt == 1:
            assert (u, v) in selected_edges
        elif gt == 0:
            assert (u, v) not in selected_edges
        elif gt is None:
            pass
        else:
            raise ValueError(f"Ground truth {gt} for edge ({u},{v}) not valid.")


def create_noise_solver(graph):
    solver = motile.Solver(graph)

    solver.add_constraint(MaxParents(1))
    solver.add_constraint(MaxChildren(1))

    solver.add_cost(NodeSelection(weight=1.0, attribute="score", constant=-10.0))
    solver.add_cost(EdgeSelection(weight=10.0, attribute="score", constant=3.0))
    solver.add_cost(Appear(constant=10.0))

    solver.add_cost(
        NodeSelection(
            weight=2,
            constant=4,
            attribute="noise",
        ),
        name="NodeNoise",
    )
    solver.add_cost(
        EdgeSelection(
            weight=6,
            constant=8,
            attribute="noise",
        ),
        name="EdgeNoise",
    )

    logger.debug("====== Initial Weights ======")
    logger.debug(solver.weights)

    return solver


def test_structsvm_noise():
    graph = create_ssvm_noise_trackgraph()
    solver = create_noise_solver(graph)

    logger.debug("====== Initial Solution ======")
    solver.solve()
    logger.debug(solver.get_variables(EdgeSelected))

    # Structured Learning
    solver.fit_weights(
        gt_attribute="gt", regularizer_weight=0.1, max_iterations=100, eps=1e-6
    )

    logger.debug("====== Learnt Weights ======")
    logger.debug(solver.weights)
    solution = solver.solve()

    logger.debug("====== Final Solution ======")
    optimal_weights = solver.weights
    logger.debug(solver.get_variables(NodeSelected))
    logger.debug(solver.get_variables(EdgeSelected))
    logger.debug(solver.features.to_ndarray())

    np.testing.assert_allclose(
        optimal_weights[("NodeSelection", "weight")], -2.77, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("NodeSelection", "constant")], 0.39, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("EdgeSelection", "weight")], -3.33, atol=0.01
    )
    np.testing.assert_allclose(
        optimal_weights[("EdgeSelection", "constant")], 0, atol=0.01
    )
    np.testing.assert_allclose(optimal_weights[("Appear", "constant")], 0.39, atol=0.01)

    def _assert_edges(solver, solution):
        edge_indicators = solver.get_variables(EdgeSelected)
        selected_edges = [
            edge for edge, index in edge_indicators.items() if solution[index] > 0.5
        ]
        for (u, v), attrs in graph.edges.items():
            gt = attrs.get("gt", None)
            if gt == 1:
                assert (u, v) in selected_edges
            elif gt == 0:
                assert (u, v) not in selected_edges
            elif gt is None:
                pass
            else:
                raise ValueError(f"Ground truth {gt} for edge ({u},{v}) not valid.")

    _assert_edges(solver, solution)

    solver = create_noise_solver(graph)
    solver.weights.from_ndarray(optimal_weights.to_ndarray())

    logger.debug("====== Learnt Weights in new solver ======")
    solution = solver.solve()
    logger.debug(solver.get_variables(NodeSelected))
    logger.debug(solver.get_variables(EdgeSelected))

    _assert_edges(solver, solution)



from collections import Counter
from unittest.mock import Mock

import motile
import networkx as nx
import pytest
from motile.constraints import MaxChildren, MaxParents
from motile.costs import (
    Appear,
    Disappear,
    EdgeDistance,
    EdgeSelection,
    NodeSelection,
    Split,
)
from motile.variables import EdgeSelected, NodeSelected


def _selected_nodes(solver: motile.Solver) -> list:
    node_indicators = solver.get_variables(NodeSelected)
    solution = solver.solve()
    return sorted([n for n, i in node_indicators.items() if solution[i] > 0.5])


def _selected_edges(solver: motile.Solver) -> list:
    edge_indicators = solver.get_variables(EdgeSelected)
    solution = solver.solve()
    return sorted([e for e, i in edge_indicators.items() if solution[i] > 0.5])


def test_graph_creation_with_hyperedges(toy_hypergraph):
    graph = toy_hypergraph
    assert len(graph.nodes) == 7
    assert len(graph.edges) == 10


def test_graph_creation_from_multiple_nx_graphs(toy_hypergraph_nx, arlo_graph_nx):
    g1 = toy_hypergraph_nx
    g2 = arlo_graph_nx
    graph = motile.TrackGraph()

    graph.add_from_nx_graph(g1)
    assert len(graph.nodes) == 7
    assert len(graph.edges) == 10
    assert graph.nodes[6]["x"] == 35
    assert "prediction_distance" not in graph.edges[(0, 2)]

    graph.add_from_nx_graph(g2)
    assert len(graph.nodes) == 7
    assert len(graph.edges) == 11
    assert graph.nodes[6]["x"] == 200
    assert "prediction_distance" in graph.edges[(0, 2)]


def test_graph_to_nx_with_hypernodes(toy_hypergraph_nx: nx.DiGraph):
    track_graph = motile.TrackGraph(nx_graph=toy_hypergraph_nx)
    nx_graph = track_graph.to_nx_graph(flatten_hyperedges=False)
    # data is slightly different so we just ensure components are equal
    nx.relabel_nodes(nx_graph, {"0_2_3": 10}, copy=False)
    assert Counter(toy_hypergraph_nx.nodes) == Counter(nx_graph.nodes)
    assert Counter(toy_hypergraph_nx.edges) == Counter(nx_graph.edges)


def test_graph_to_nx_flattened(toy_hypergraph_nx: nx.DiGraph):
    track_graph = motile.TrackGraph(nx_graph=toy_hypergraph_nx)
    nx_graph = track_graph.to_nx_graph(flatten_hyperedges=True)
    # data is slightly different so we just ensure components are equal
    expected_graph = toy_hypergraph_nx.copy()
    expected_graph.remove_node(10)
    expected_graph.add_edge(0, 2)
    expected_graph.add_edge(0, 3)
    assert Counter(expected_graph.nodes) == Counter(nx_graph.nodes)
    assert Counter(expected_graph.edges) == Counter(nx_graph.edges)


def test_graph_creation_wrong_frame_attr(toy_hypergraph_nx):
    with pytest.raises(KeyError):
        motile.TrackGraph(toy_hypergraph_nx, frame_attribute="time")


def test_solver(arlo_graph):
    graph = arlo_graph

    solver = motile.Solver(graph)
    solver.add_cost(NodeSelection(weight=-1.0, attribute="score", constant=-100.0))
    solver.add_cost(
        EdgeSelection(weight=0.5, attribute="prediction_distance", constant=-1.0)
    )
    solver.add_cost(EdgeDistance(position_attribute=("x",), weight=0.5))
    solver.add_cost(Appear(constant=200.0, attribute="score", weight=-1.0))
    solver.add_cost(Disappear(constant=55.0))
    solver.add_cost(Split(constant=100.0, attribute="score", weight=1.0))

    solver.add_constraint(MaxParents(1))
    solver.add_constraint(MaxChildren(2))

    callback = Mock()
    solution = solver.solve(on_event=callback)
    assert callback.call_count
    subgraph = solver.get_selected_subgraph()

    assert (
        list(subgraph.edges)
        == _selected_edges(solver)
        == [(0, 2), (1, 3), (2, 4), (3, 5)]
    )
    assert list(subgraph.nodes) == _selected_nodes(solver) == [0, 1, 2, 3, 4, 5]
    cost = solution.get_value()
    assert cost == -604.0, f"{cost=}"



from __future__ import annotations

from typing import TYPE_CHECKING, Collection

from .variable import Variable

if TYPE_CHECKING:
    from motile._types import Node
    from motile.solver import Solver


class NodeSelected(Variable["Node"]):
    """Binary variable indicating whether a node is part of the solution or not."""

    @staticmethod
    def instantiate(solver: Solver) -> Collection[Node]:
        return solver.graph.nodes



from .edge_selected import EdgeSelected
from .node_appear import NodeAppear
from .node_disappear import NodeDisappear
from .node_merge import NodeMerge
from .node_selected import NodeSelected
from .node_split import NodeSplit
from .variable import Variable

__all__ = [
    "EdgeSelected",
    "NodeAppear",
    "NodeDisappear",
    "NodeMerge",
    "NodeSelected",
    "NodeSplit",
    "Variable",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import ilpy
from ilpy.expressions import Constant

from ..variables import EdgeSelected
from .constraint import Constraint

if TYPE_CHECKING:
    from motile.solver import Solver


class MaxChildren(Constraint):
    r"""Ensures that every selected node has no more than ``max_children``.

    Where a "child" is a selected edges to the next frame.

    Adds the following linear constraint for each node :math:`v`:

    .. math::

      \sum_{e \in \text{out_edges}(v)} x_e \leq \text{max_children}

    Args:
        max_children:
            The maximum number of children allowed.
    """

    def __init__(self, max_children: int) -> None:
        self.max_children = max_children

    def instantiate(self, solver: Solver) -> Iterable[ilpy.Expression]:
        edge_indicators = solver.get_variables(EdgeSelected)

        for node in solver.graph.nodes:
            n_edges = sum(
                (edge_indicators[e] for e in solver.graph.next_edges[node]), Constant(0)
            )

            yield n_edges <= self.max_children



import motile
import networkx as nx
from motile.costs import EdgeSelection, NodeSelection


def test_empty_graph() -> None:
    """Test that solving an empty graph does not error and returns empty solution."""
    nx_graph = nx.DiGraph()
    graph = motile.TrackGraph(nx_graph)

    solver = motile.Solver(graph)
    solver.add_cost(NodeSelection(constant=-1))
    solver.add_cost(EdgeSelection(constant=-1))

    # Should not error
    solver.solve()
    solution_graph = solver.get_selected_subgraph()

    # Solution should be empty
    assert len(solution_graph.nodes) == 0
    assert len(solution_graph.edges) == 0


def test_graph_with_no_edges() -> None:
    """Test that solving a graph with nodes but no edges does not error."""
    cells = [
        {"id": 0, "t": 0},
        {"id": 1, "t": 0},
        {"id": 2, "t": 1},
    ]

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    # No edges added

    graph = motile.TrackGraph(nx_graph)

    solver = motile.Solver(graph)
    solver.add_cost(NodeSelection(constant=-1))
    solver.add_cost(EdgeSelection(constant=-1))

    # Should not error
    solver.solve()
    solution_graph = solver.get_selected_subgraph()

    # All nodes should be selected due to negative cost
    assert set(solution_graph.nodes.keys()) == {0, 1, 2}
    # No edges should be selected (none exist)
    assert len(solution_graph.edges) == 0



import motile
import pytest
from motile.constraints import (
    ExclusiveNodes,
    ExpressionConstraint,
    MaxChildren,
    MaxParents,
    Pin,
)
from motile.costs import Appear, EdgeSelection, NodeSelection

from .test_api import _selected_edges, _selected_nodes


@pytest.fixture
def solver(arlo_graph):
    return motile.Solver(arlo_graph)


def test_graph_casting(arlo_graph_nx) -> None:
    with pytest.warns(UserWarning, match="Coercing networkx graph to TrackGraph"):
        motile.Solver(arlo_graph_nx)


def test_pin(solver: motile.Solver) -> None:
    # pin the value of two edges:
    solver.graph.edges[(0, 2)]["pin_to"] = False  # type: ignore
    solver.graph.edges[(3, 6)]["pin_to"] = True  # type: ignore

    assert _selected_edges(solver) != [(3, 6)], "test invalid"
    solver.add_constraint(Pin("pin_to"))
    assert _selected_edges(solver) == [(3, 6)]


def test_expression(solver: motile.Solver) -> None:
    solver.graph.nodes[5]["color"] = "red"  # type: ignore
    solver.add_cost(NodeSelection(weight=-1.0, attribute="score", constant=-1))

    assert _selected_nodes(solver) != [1, 6], "test invalid"
    # constrain solver based on attributes of nodes/edges
    expr = "x > 140 and t != 1 and color != 'red'"
    solver.add_constraint(ExpressionConstraint(expr))
    assert _selected_nodes(solver) == [1, 6]


def test_max_children(solver: motile.Solver) -> None:
    solver.add_cost(
        EdgeSelection(weight=1.0, attribute="prediction_distance", constant=-100)
    )

    expect = [(0, 2), (1, 3), (2, 4), (3, 5)]
    assert _selected_edges(solver) != expect, "test invalid"
    solver.add_constraint(MaxChildren(1))
    assert _selected_edges(solver) == expect


def test_max_parents(solver: motile.Solver) -> None:
    solver.add_cost(
        EdgeSelection(weight=1.0, attribute="prediction_distance", constant=-100)
    )

    expect = [(0, 2), (1, 3), (2, 4), (3, 5), (3, 6)]
    assert _selected_edges(solver) != expect, "test invalid"
    solver.add_constraint(MaxParents(1))
    assert _selected_edges(solver) == expect


def test_exlusive_nodes(solver: motile.Solver) -> None:
    exclusive_sets = [
        [0, 1],
        [2, 3],
        [4, 5],
    ]

    solver.add_cost(
        EdgeSelection(weight=1.0, attribute="prediction_distance", constant=-100)
    )
    solver.add_cost(Appear(constant=2.0, attribute="score", weight=0))
    solver.add_constraint(MaxParents(1))
    solver.add_constraint(MaxChildren(2))
    solver.add_constraint(ExclusiveNodes(exclusive_sets))

    assert _selected_nodes(solver) == [1, 3, 5, 6], "test invalid"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    import ilpy

    from motile.solver import Solver


class Constraint(ABC):
    """A base class for a constraint that can be added to a solver."""

    @abstractmethod
    def instantiate(
        self, solver: Solver
    ) -> Iterable[ilpy.Constraint | ilpy.Expression]:
        """Create and return specific linear constraints for the given solver.

        Args:
            solver:
                The :class:`~motile.Solver` instance to create linear constraints for.

        Returns:
            An iterable of :class:`ilpy.Constraint`.
        """



from __future__ import annotations

from typing import TYPE_CHECKING, Collection, Iterable

import ilpy

from .edge_selected import EdgeSelected
from .variable import Variable

if TYPE_CHECKING:
    from motile._types import Node
    from motile.solver import Solver


class NodeSplit(Variable):
    r"""Binary variable indicating whether a node has more than one child.

    (i.e., the node is selected and has more than one selected outgoing edge).

    This variable is coupled to the edge selection variables through the
    following linear constraints:

    .. math::

        2 s_v\; - &\sum_{e\in\text{out_edges}(v)} x_e &\leq&\;\; 0

        (|\text{out_edges}(v)| - 1) s_v\; - &\sum_{e\in\text{out_edges}(v)}
        x_e &\geq&\;\; -1

    where :math:`x_e` are selection indicators for edge :math:`e`, and
    :math:`s_v` is the split indicator for node :math:`v`.
    """

    @staticmethod
    def instantiate(solver: Solver) -> Collection[Node]:
        return solver.graph.nodes

    @staticmethod
    def instantiate_constraints(solver: Solver) -> Iterable[ilpy.Constraint]:
        split_indicators = solver.get_variables(NodeSplit)
        edge_indicators = solver.get_variables(EdgeSelected)

        for node in solver.graph.nodes:
            next_edges = solver.graph.next_edges[node]

            # Ensure that the following holds:
            #
            # split = 0 <=> sum(next_selected) <= 1
            # split = 1 <=> sum(next_selected) > 1
            #
            # Two linear constraints are needed for that:
            #
            # (1) 2 * split - sum(next_selected) <= 0
            # (2) (num_next - 1) * split - sum(next_selected) >= -1

            constraint1 = ilpy.Constraint()
            constraint2 = ilpy.Constraint()

            constraint1.set_coefficient(split_indicators[node], 2.0)
            constraint2.set_coefficient(split_indicators[node], len(next_edges) - 1.0)

            for next_edge in next_edges:
                constraint1.set_coefficient(edge_indicators[next_edge], -1.0)
                constraint2.set_coefficient(edge_indicators[next_edge], -1.0)

            constraint1.set_relation(ilpy.Relation.LessEqual)
            constraint2.set_relation(ilpy.Relation.GreaterEqual)

            constraint1.set_value(0.0)
            constraint2.set_value(-1.0)

            yield constraint1
            yield constraint2



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import structsvm as ssvm

from .variables import EdgeSelected, NodeSelected

if TYPE_CHECKING:
    from motile.solver import Solver


def fit_weights(
    solver: Solver,
    gt_attribute: str,
    regularizer_weight: float,
    max_iterations: int | None,
    eps: float,
) -> np.ndarray:
    """Return the optimal weights for the given solver.

    This uses `structsvm.BundleMethod` to fit the weights.

    Args:
        solver (Solver):
            The solver to fit the weights for.
        gt_attribute (str):
            Node/edge attribute that marks the ground truth for fitting.
            `gt_attribute` is expected to be set to `1` for objects labeled as
            ground truth, `0` for objects explicitly labeled as not part of the
            ground truth, and `None` or not set for unlabeled objects.
        regularizer_weight (float):
            The weight of the quadratic regularizer.
        max_iterations (int):
            Maximum number of gradient steps in the structured SVM.
        eps (float):
            Convergence threshold.

    Returns:
        np.ndarray:
            The optimal weights for the given solver.
    """
    features = solver.features.to_ndarray()

    mask = np.zeros((solver.num_variables,), dtype=np.float32)
    ground_truth = np.zeros((solver.num_variables,), dtype=np.float32)

    for node, index in solver.get_variables(NodeSelected).items():
        gt = solver.graph.nodes[node].get(gt_attribute, None)
        if gt is not None:
            mask[index] = 1.0
            ground_truth[index] = gt

    for edge, index in solver.get_variables(EdgeSelected).items():
        gt = solver.graph.edges[edge].get(gt_attribute, None)
        if gt is not None:
            mask[index] = 1.0
            ground_truth[index] = gt

    loss = ssvm.SoftMarginLoss(
        solver.constraints,
        features.T,  # TODO: fix in ssvm
        ground_truth,
        ssvm.HammingCosts(ground_truth, mask),
    )
    bundle_method = ssvm.BundleMethod(
        loss.value_and_gradient,
        dims=features.shape[1],
        regularizer_weight=regularizer_weight,
        eps=eps,
    )

    return bundle_method.optimize(max_iterations)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
from datetime import datetime

import motile
import tomli

with open("../../pyproject.toml", "rb") as fh:
    project = tomli.load(fh)["project"]

author_list = ", ".join([author["name"] for author in project["authors"]])

project = "motile"
copyright = f"{datetime.now().year}, {author_list}, Ana Cristina Pascual Ramos"
author = author_list
version = motile.__version__
release = motile.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "jupyter_sphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.githubpages",
    "sphinx.ext.mathjax",
    "sphinx_rtd_theme",
    "sphinx_togglebutton",
    "sphinxcontrib.jquery",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_logo = "img/motile.svg"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_show_sourcelink = False

# Version switcher for versioned docs on GitHub Pages.
# The DOCS_VERSION env var is set by CI; defaults to the package version locally.
docs_version = os.environ.get("DOCS_VERSION", f"v{version}")
html_context = {
    "current_version": docs_version,
    "versions": [
        ("stable", "/motile/stable/"),
        ("main", "/motile/main/"),
    ],
}

togglebutton_hint = ""
togglebutton_hint_hide = ""

pygments_style = "lovelace"

# Napoleon settings
napoleon_google_docstring = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "networkx": ("https://networkx.org/documentation/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}



from typing import Any, Callable, List

Callback = Callable[[float, float], Any]


class Weight:
    """A single Weight with observer/callback pattern on update.

    See also :class:`motile.costs.weights.Weights`.

    Args:
        initial_value:
            The initial value of the weight.
    """

    def __init__(self, initial_value: float) -> None:
        self._value = initial_value
        self._modify_callbacks: List[Callback] = []

    @property
    def value(self) -> float:
        """Return the value of this weight."""
        return self._value

    @value.setter
    def value(self, new_value: float) -> None:
        """Set the value of this weight."""
        old_value = self._value
        self._value = new_value
        self._notify_modified(old_value, new_value)

    def register_modify_callback(self, callback: Callback) -> None:
        """Register a ``callback`` to be called when the weight is modified.

        Args:
            callback:
                A function that takes two arguments: the old value and the new value.
        """
        self._modify_callbacks.append(callback)

    def _notify_modified(self, old_value: float, new_value: float) -> None:
        for callback in self._modify_callbacks:
            callback(old_value, new_value)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import ilpy
from ilpy.expressions import Constant

from ..variables import EdgeSelected
from .constraint import Constraint

if TYPE_CHECKING:
    from motile.solver import Solver


class MaxParents(Constraint):
    r"""Ensures that every selected node has no more than ``max_parents``.

    Where a "parent" is defined as an incoming selected edge from the previous frame.

    Adds the following linear constraint for each node :math:`v`:

    .. math::

      \sum_{e \in \text{in_edges}(v)} x_e \leq \text{max_parents}

    Args:
        max_parents:
            The maximum number of parents allowed.
    """

    def __init__(self, max_parents: int) -> None:
        self.max_parents = max_parents

    def instantiate(self, solver: Solver) -> Iterable[ilpy.Expression]:
        edge_indicators = solver.get_variables(EdgeSelected)

        for node in solver.graph.nodes:
            # all incoming edges
            s = sum(
                (edge_indicators[e] for e in solver.graph.prev_edges[node]),
                start=Constant(0),
            )
            yield s <= self.max_parents



from __future__ import annotations

from typing import TYPE_CHECKING

from ..variables import NodeAppear
from .cost import Cost
from .weight import Weight

if TYPE_CHECKING:
    from motile.solver import Solver


class Appear(Cost):
    """Cost for :class:`~motile.variables.NodeAppear` variables.

    This is cost is not applied to nodes in the first frame of the graph.

    Args:
        weight:
            The weight to apply to the cost attribute of each starting track.
            Defaults to 1.

        attribute:
            The name of the attribute to use to look up cost. Default is
            ``None``, which means that a constant cost is used.

        constant:
            A constant cost for each node that starts a track. Defaults to 0.

        ignore_attribute:
            The name of an optional node attribute that, if it is set and
            evaluates to ``True``, will not set the appear cost for that node.
            Defaults to None
    """

    def __init__(
        self,
        weight: float = 1,
        attribute: str | None = None,
        constant: float = 0,
        ignore_attribute: str | None = None,
    ) -> None:
        self.weight = Weight(weight)
        self.constant = Weight(constant)
        self.attribute = attribute
        self.ignore_attribute = ignore_attribute

    def apply(self, solver: Solver) -> None:
        appear_indicators = solver.get_variables(NodeAppear)
        G = solver.graph

        for node, index in appear_indicators.items():
            if self.ignore_attribute is not None:
                if G.nodes[node].get(self.ignore_attribute, False):
                    continue
            if G.nodes[node][G.frame_attribute] == G.get_frames()[0]:
                continue
            if self.attribute is not None:
                solver.add_variable_cost(
                    index, G.nodes[node][self.attribute], self.weight
                )
            solver.add_variable_cost(index, 1.0, self.constant)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np

from ..variables import EdgeSelected
from .cost import Cost
from .weight import Weight

if TYPE_CHECKING:
    import networkx as nx

    from motile.solver import Solver


class EdgeDistance(Cost):
    """Cost for :class:`~motile.variables.EdgeSelected` variables.

    Cost is based on the spatial distance of the incident nodes.

    Args:
        position_attribute:
            The name of the node attribute that corresponds to the spatial
            position. Can also be given as a tuple of multiple coordinates,
            e.g., ``('z', 'y', 'x')``.

        weight:
            The weight to apply to the distance to convert it into a cost.

        constant:
            A constant cost for each selected node. Default is ``0.0``.
    """

    def __init__(
        self,
        position_attribute: str | tuple[str, ...],
        weight: float = 1.0,
        constant: float = 0.0,
    ) -> None:
        self.position_attribute = position_attribute
        self.weight = Weight(weight)
        self.constant = Weight(constant)

    def apply(self, solver: Solver) -> None:
        edge_variables = solver.get_variables(EdgeSelected)
        for key, index in edge_variables.items():
            u, v = cast("tuple[int, int]", key)
            pos_u = self.__get_node_position(solver.graph, u)
            pos_v = self.__get_node_position(solver.graph, v)

            feature = np.linalg.norm(pos_u - pos_v)

            solver.add_variable_cost(index, feature, self.weight)
            solver.add_variable_cost(index, 1.0, self.constant)

    def __get_node_position(self, graph: nx.DiGraph, node: int) -> np.ndarray:
        if isinstance(self.position_attribute, tuple):
            return np.array([graph.nodes[node][p] for p in self.position_attribute])
        else:
            return np.array(graph.nodes[node][self.position_attribute])



import motile
import networkx as nx
from motile.costs import (
    Appear,
    Disappear,
    EdgeSelection,
    Merge,
    NodeSelection,
)


def test_appear_cost(arlo_graph):
    solver = motile.Solver(arlo_graph)

    # Make a slightly negative node cost, and a very positive appear cost and edge
    # cost. We expect only selecting nodes in the first frame, where by default the
    # appear cost is ignored, and not selecting any edges
    solver.add_cost(NodeSelection(weight=0, attribute="score", constant=-1))
    solver.add_cost(Appear(constant=100))
    solver.add_cost(
        EdgeSelection(weight=0, attribute="prediction_distance", constant=100)
    )
    solver.solve()
    solution_graph = solver.get_selected_subgraph()
    assert list(solution_graph.nodes.keys()) == [0, 1]
    assert len(solution_graph.edges) == 0

    ignore_attr = "ignore"
    # now also ignore the appear cost in the second frame
    for second_node in arlo_graph.nodes_by_frame(1):
        arlo_graph.nodes[second_node][ignore_attr] = True
    # but not the third frame
    for third_node in arlo_graph.nodes_by_frame(2):
        arlo_graph.nodes[third_node][ignore_attr] = False

    solver = motile.Solver(arlo_graph)

    # Resolving should also select nodes in second frame
    solver.add_cost(NodeSelection(weight=0, attribute="score", constant=-1))
    solver.add_cost(Appear(constant=100, ignore_attribute=ignore_attr))
    solver.add_cost(
        EdgeSelection(weight=0, attribute="prediction_distance", constant=100)
    )
    solver.solve()
    solution_graph = solver.get_selected_subgraph()
    assert list(solution_graph.nodes.keys()) == [0, 1, 2, 3]
    assert len(solution_graph.edges) == 0


def test_disappear_cost(arlo_graph):
    solver = motile.Solver(arlo_graph)

    # make a slightly negative node cost, and a positive disappear cost and edge cost
    # we expect only selecting nodes in the last frame, where by default the disappear
    # cost is ignored, and not selecting any edges
    solver.add_cost(NodeSelection(weight=0, attribute="score", constant=-1))
    solver.add_cost(Disappear(constant=100))
    solver.add_cost(
        EdgeSelection(weight=0, attribute="prediction_distance", constant=100)
    )
    solver.solve()
    solution_graph = solver.get_selected_subgraph()
    assert list(solution_graph.nodes.keys()) == [4, 5, 6]
    assert len(solution_graph.edges) == 0

    ignore_attr = "ignore"
    # now also ignore the disappear cost in the second frame
    for second_node in arlo_graph.nodes_by_frame(1):
        arlo_graph.nodes[second_node][ignore_attr] = True
    # but not the first frame
    for first_node in arlo_graph.nodes_by_frame(0):
        arlo_graph.nodes[first_node][ignore_attr] = False

    solver = motile.Solver(arlo_graph)

    # Resolving should also select nodes in second frame
    solver.add_cost(NodeSelection(weight=0, attribute="score", constant=-1))
    solver.add_cost(Disappear(constant=100, ignore_attribute=ignore_attr))
    solver.add_cost(
        EdgeSelection(weight=0, attribute="prediction_distance", constant=100)
    )
    solver.solve()
    solution_graph = solver.get_selected_subgraph()
    assert list(solution_graph.nodes.keys()) == [2, 3, 4, 5, 6]
    assert len(solution_graph.edges) == 0


def test_constant_merge_cost() -> None:
    """Test that merge cost prevents merges when applied.

    Graph structure:
        t=0: node 0, node 1
        t=1: node 2
        edges: 0->2, 1->2

    With only negative edge selection cost, both edges should be selected
    (resulting in a merge). Adding a merge cost should prevent the merge,
    resulting in only one edge being selected.
    """
    # Create nodes
    cells = [
        {"id": 0, "t": 0},
        {"id": 1, "t": 0},
        {"id": 2, "t": 1},
    ]

    # Create edges (both leading to node 2, creating potential merge)
    edges = [
        {"source": 0, "target": 2},
        {"source": 1, "target": 2},
    ]

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    nx_graph.add_edges_from([(edge["source"], edge["target"], edge) for edge in edges])

    graph = motile.TrackGraph(nx_graph)

    # First test: without merge cost, both edges should be selected
    solver = motile.Solver(graph)
    solver.add_cost(EdgeSelection(constant=-1.0))
    solver.solve()
    solution_graph = solver.get_selected_subgraph().to_nx_graph()

    # Should select all nodes and both edges (merge occurs)
    assert set(solution_graph.nodes.keys()) == {0, 1, 2}
    assert len(solution_graph.edges) == 2
    assert solution_graph.has_edge(0, 2)
    assert solution_graph.has_edge(1, 2)

    # Second test: with merge cost, only one edge should be selected
    solver = motile.Solver(graph)
    solver.add_cost(EdgeSelection(constant=-1.0))
    solver.add_cost(Merge(constant=10.0))  # High cost to prevent merge
    solver.solve()
    solution_graph = solver.get_selected_subgraph().to_nx_graph()

    # Should select all nodes but only one edge (no merge)
    assert set(solution_graph.nodes.keys()) == {0, 1, 2}
    assert len(solution_graph.edges) == 1
    # Either edge 0->2 or 1->2 should be selected, but not both
    assert solution_graph.has_edge(0, 2) or solution_graph.has_edge(1, 2)
    assert not (solution_graph.has_edge(0, 2) and solution_graph.has_edge(1, 2))


def test_variable_merge_cost() -> None:
    """Test that merge cost can use node attributes to selectively allow merges.

    Graph structure:
        t=0: node 0, node 1, node 2
        t=1: node 3 (merge_cost=-1.0), node 4 (merge_cost=5.0)
        edges: 0->3, 1->3, 1->4, 2->4

    With negative edge selection cost and attribute-based merge cost,
    only the node with negative merge_cost should have a merge (node 3).
    Node 4 with positive merge_cost should not have a merge.
    """
    # Create nodes - all nodes need merge_cost attribute
    cells = [
        {"id": 0, "t": 0, "merge_cost": 0.0},
        {"id": 1, "t": 0, "merge_cost": 0.0},
        {"id": 2, "t": 0, "merge_cost": 0.0},
        {"id": 3, "t": 1, "merge_cost": -1.0},  # negative cost = good merge
        {"id": 4, "t": 1, "merge_cost": 5.0},  # positive cost = bad merge
    ]

    # Create edges - two edges to each node in time 1 (potential merges)
    edges = [
        {"source": 0, "target": 3},
        {"source": 1, "target": 3},
        {"source": 1, "target": 4},
        {"source": 2, "target": 4},
    ]

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from([(cell["id"], cell) for cell in cells])
    nx_graph.add_edges_from([(edge["source"], edge["target"], edge) for edge in edges])

    graph = motile.TrackGraph(nx_graph)

    solver = motile.Solver(graph)
    solver.add_cost(EdgeSelection(constant=-1.0))
    solver.add_cost(Merge(attribute="merge_cost", weight=1.0, constant=0.0))
    solver.solve()
    solution_graph = solver.get_selected_subgraph().to_nx_graph()

    # Should select all nodes
    assert set(solution_graph.nodes.keys()) == {0, 1, 2, 3, 4}

    # Node 3 should have a merge (both edges 0->3 and 1->3 selected)
    # because its merge_cost is negative (-1.0), making the total cost attractive
    assert solution_graph.has_edge(0, 3)
    assert solution_graph.has_edge(1, 3)

    # Node 4 should NOT have a merge (only one edge selected)
    # because its merge_cost is positive (5.0), making the merge too expensive
    edges_to_4 = [
        solution_graph.has_edge(1, 4),
        solution_graph.has_edge(2, 4),
    ]
    assert sum(edges_to_4) == 1  # exactly one edge to node 4



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Hashable, Iterable, Optional

import numpy as np

if TYPE_CHECKING:
    from .weight import Weight

Callback = Callable[[Optional[float], float], Any]


class Weights:
    """A simple container for weights with observer/callback pattern on update.

    A :class:`motile.Solver` has a :class:`Weights` instance that is used to
    store the weights of the model.  Changes to the weights can be observed with
    ``Solver.weights.register_modify_callback``
    """

    def __init__(self) -> None:
        self._weights: list[Weight] = []
        self._weights_by_name: dict[Hashable, Weight] = {}
        self._weight_indices: dict[Weight, int] = {}
        self._modify_callbacks: list[Callback] = []

    def add_weight(self, weight: Weight, name: Hashable) -> int:
        """Add a weight to the container.

        Args:
            weight:
                The :class:`~motile.costs.Weight` to add.
            name:
                The name of the weight.

        Returns:
            int: the index of the weight
        """
        weight_index = len(self._weights)
        self._weight_indices[weight] = weight_index
        self._weights.append(weight)
        self._weights_by_name[name] = weight

        for callback in self._modify_callbacks:
            weight.register_modify_callback(callback)

        self._notify_modified(None, weight.value)
        return weight_index

    def register_modify_callback(self, callback: Callback) -> None:
        """Register ``callback`` to be called when a weight is modified.

        Args:
            callback:
                A function that takes two arguments: the old value (which may be
                ``None``) and the new value.
        """
        self._modify_callbacks.append(callback)
        for weight in self._weights:
            weight.register_modify_callback(callback)

    def to_ndarray(self) -> np.ndarray:
        """Export the weights as a numpy array.

        Note: you can also use ``np.asarray(weights)``.
        """
        return np.array([w.value for w in self._weights], dtype=np.float32)

    def __array__(self) -> np.ndarray:
        return self.to_ndarray()

    def from_ndarray(self, values: Iterable[float]) -> None:
        """Update weights from an iterable of floats."""
        for weight, value in zip(self._weights, values):
            weight.value = value

    def index_of(self, weight: Weight) -> int:
        """Return the index of ``weight`` in this container."""
        return self._weight_indices[weight]

    def __getitem__(self, name: str) -> float:
        """Return the value of the weight with the given name."""
        return self._weights_by_name[name].value

    def __setitem__(self, name: str, value: float) -> None:
        """Set the value of the weight with the given name."""
        self._weights_by_name[name].value = value

    def _notify_modified(self, old_value: float | None, new_value: float) -> None:
        for callback in self._modify_callbacks:
            callback(old_value, new_value)

    def __repr__(self) -> str:
        return "".join(
            f"{name} = {weight.value}\n"
            for name, weight in self._weights_by_name.items()
        )



from __future__ import annotations

import ast
import contextlib
from typing import TYPE_CHECKING, Union

import ilpy

from ..variables import EdgeSelected, NodeSelected, Variable
from .constraint import Constraint

if TYPE_CHECKING:
    from motile._types import Attributes, GenericEdge, Node
    from motile.solver import Solver

    NodesOrEdges = Union[dict[Node, Attributes], dict[GenericEdge, Attributes]]


class ExpressionConstraint(Constraint):
    """Enforce selection of nodes/edges based on an expression.

    The expression string is evaluated with the node/edge dict as a namespace.

    This is a powerful general constraint that allows you to select nodes/edges based on
    any combination of node/edge attributes. The `expression` string is evaluated for
    each node/edge (assuming eval_nodes/eval_edges is True) using the actual node object
    as a namespace to populate any variables names used in the provided expression. If
    the expression evaluates to True, the node/edge is selected; otherwise, it is
    excluded.

    This takes advantaged of python's `eval` function, like this:

    .. code-block:: python

        my_expression = "some_attribute == True"
        eval(my_expression, None, {"some_attribute": True})  # returns True (select)
        eval(my_expression, None, {"some_attribute": False})  # returns False (exclude)
        eval(my_expression, None, {})  # raises NameError (do nothing)

    Args:
        expression:
            An expression to evaluate for each node/edge. The expression must
            evaluate to a boolean value. The expression can use any names of
            node/edge attributes as variables.
        eval_nodes:
            Whether to evaluate the expression for nodes. By default, True.
        eval_edges:
            Whether to evaluate the expression for edges. By default, True.

    Example:
        If the nodes of a graph are:

        >>> cells = [
        ...     {"id": 0, "t": 0, "color": "red", "score": 1.0},
        ...     {"id": 1, "t": 0, "color": "green", "score": 1.0},
        ...     {"id": 2, "t": 1, "color": "blue", "score": 1.0},
        ... ]

        Then the following constraint will select node 0:

        >>> expr = "t == 0 and color != 'green'"
        >>> solver.add_constraints(ExpressionConstraint(expr))
    """

    def __init__(
        self, expression: str, eval_nodes: bool = True, eval_edges: bool = True
    ) -> None:
        try:
            tree = ast.parse(expression, mode="eval")
            if not isinstance(tree, ast.Expression):
                raise SyntaxError
        except SyntaxError:
            raise ValueError(f"Invalid expression: {expression}") from None

        self._expression = compile(expression, "<string>", "eval")
        self.eval_nodes = eval_nodes
        self.eval_edges = eval_edges

    def instantiate(self, solver: Solver) -> list[ilpy.Constraint]:
        # create two constraints: one to select nodes/edges, and one to exclude
        select = ilpy.Constraint()
        exclude = ilpy.Constraint()
        n_selected = 0  # number of nodes/edges selected

        to_evaluate: list[tuple[NodesOrEdges, type[Variable]]] = []
        if self.eval_nodes:
            to_evaluate.append((solver.graph.nodes, NodeSelected))
        if self.eval_edges:
            to_evaluate.append((solver.graph.edges, EdgeSelected))

        for nodes_or_edges, VariableType in to_evaluate:
            indicator_variables = solver.get_variables(VariableType)
            for id_, node_or_edge in nodes_or_edges.items():
                with contextlib.suppress(NameError):
                    # Here is where the expression string is evaluated.
                    # We use the node/edge dict as a namespace to look up variables.
                    # if the expression uses a variable name that is not in the dict,
                    # a NameError will be raised.
                    # contextlib.suppress (above) will just skip it and move on...
                    if eval(self._expression, None, node_or_edge):
                        # if the expression evaluates to True, we select the node/edge
                        select.set_coefficient(indicator_variables[id_], 1)
                        n_selected += 1
                    else:
                        # Otherwise, we exclude it.
                        exclude.set_coefficient(indicator_variables[id_], 1)

        # finally, apply the relation and value to the constraints
        select.set_relation(ilpy.Relation.Equal)
        select.set_value(n_selected)

        exclude.set_relation(ilpy.Relation.Equal)
        exclude.set_value(0)

        return [select, exclude]



from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Any, DefaultDict, Hashable, Iterator, cast

import networkx

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from typing import Mapping

    from typing_extensions import TypeGuard

    from motile._types import (
        Attributes,
        Edge,
        GenericEdge,
        HyperEdge,
        Node,
        Nodes,
    )


class TrackGraph:
    """A graph of nodes in time & space, with edges connecting them across time.

    This wraps a :class:`networkx.DiGraph` object.

    Both ``nodes`` and ``edges`` are represented by a dictionary of properties.

    Provides a few convenience methods for time series graphs in addition to
    all the methods inherited from :class:`networkx.DiGraph`.

    Args:
        nx_graph:

            A directed networkx graph representing the TrackGraph to be created.
            Hyperedges are represented by networkx nodes that do not have the
            ``frame_attribute`` and are connected to nodes that do have this
            attribute.

        frame_attribute:

            The name of the node attribute that corresponds to the frame (i.e.,
            the time dimension) of the object. Defaults to ``'t'``.
    """

    def __init__(
        self,
        nx_graph: networkx.DiGraph | None = None,
        frame_attribute: str = "t",
    ):
        self.frame_attribute = frame_attribute
        self._graph_changed = True

        self.nodes: dict[Node, Attributes] = {}
        self.edges: dict[GenericEdge, Attributes] = {}
        self.prev_edges: defaultdict[Node, list[GenericEdge]] = DefaultDict(list)
        self.next_edges: defaultdict[Node, list[GenericEdge]] = DefaultDict(list)

        if nx_graph:
            self.add_from_nx_graph(nx_graph)

        self._update_metadata()

    def add_node(self, node_id: Node, data: Attributes) -> None:
        """Adds a new node to this TrackGraph.

        Args:
            node_id: the node to be added.
            data: all properties associated to the added node.
        """
        assert self.frame_attribute in data, (
            "Nodes in the track graph need to have (at least) a frame "
            f"attribute {self.frame_attribute}"
        )
        self.nodes[node_id] = data
        self._graph_changed = True

    def add_edge(self, edge_id: GenericEdge, data: Attributes) -> None:
        """Adds an edge to this TrackGraph.

        Args:
            edge_id: an ``GenericEdge`` (tuple of Nodes) defining the edge
                (or hyperedge) to be added.
            data: all properties associated to the added edge.
        """
        self.edges[edge_id] = data

        if self.is_hyperedge(edge_id):
            us, vs = edge_id
            for v in vs:
                self.prev_edges[v].append(edge_id)
            for u in us:
                self.next_edges[v].append(edge_id)
        else:
            # normal (u, v) edge
            u, v = cast("Edge", edge_id)
            self.prev_edges[v].append(edge_id)
            self.next_edges[u].append(edge_id)

        self._graph_changed = True

    def add_from_nx_graph(self, nx_graph: networkx.DiGraph) -> None:
        """Add nodes/edges from ``nx_graph`` to this TrackGraph.

        Hyperedges are represented by nodes in the ``nx_graph`` that do not have the
        ``frame_attribute`` property. All 'regular' nodes connected to such a hyperedge
        node will be added as a hyperedge.

        Args:
            nx_graph (networkx.DiGraph):

                A directed networkx graph representing a TrackGraph to be added.
                Hyperedges are represented by networkx nodes that do not have the
                ``frame_attribute`` and are connected to nodes that do have this
                attribute.

                Duplicate nodes and edges will not be added again but new attributes
                associated to nodes and edges added. If attributes of existing nodes
                or edges do already exist, the values set in the given ``nx_graph``
                will be updating the previously set values.
        """
        nodes_count = 0
        # add all regular nodes (all but ones representing hyperedges)
        for node, data in nx_graph.nodes.items():
            if self.frame_attribute in data:
                if node not in self.nodes:
                    self.nodes[node] = data
                else:
                    self.nodes[node] |= data
                nodes_count += 1

        # graph without nodes, it's very likely this was not intentional
        if nodes_count == 0:
            raise KeyError(
                f"No nodes with `frame_attribute` '{self.frame_attribute}' found in "
                "the `nx_graph`.\nIt's likely the wrong `frame_attribute` was set."
            )

        # add all edges and hyperedges
        for (u, v), data in nx_graph.edges.items():
            # do nothing when you have an nx_edge leaving a hyperedge node
            # (we will add this hyperedge when we encounter it via an incoming edge)
            if self._is_hyperedge_nx_node(nx_graph, u):
                continue
            # add hyperedge when nx_edge leads to hyperedge node
            if self._is_hyperedge_nx_node(nx_graph, v):
                # get data off the hypernode
                data = nx_graph.nodes[v]
                edge = self._convert_nx_hypernode(nx_graph, v)
                in_nodes, out_nodes = edge
                # avoid adding duplicates
                if edge not in self.edges:
                    self.edges[edge] = data
                    for in_node in in_nodes:
                        self.next_edges[in_node].append(edge)
                    for out_node in out_nodes:
                        self.prev_edges[out_node].append(edge)
            else:  # add a regular edge otherwise
                self.edges[(u, v)] = data
                self.prev_edges[v].append((u, v))
                self.next_edges[u].append((u, v))

    def to_nx_graph(self, flatten_hyperedges: bool = True) -> networkx.DiGraph:
        """Convert a this TrackGraph into a networkx DiGraph.

        Args:
            flatten_hyperedges (bool, optional): If True, include one edge for each
                (source, target) combo in a hyperedge. If False, introduce a new
                hypernode to represent hyperedges. Defaults to True.

        Returns:
            networkx.DiGraph: Directed networkx graph with same nodes, edges, and
                attributes.
        """
        nx_graph = networkx.DiGraph()
        nodes_list = list(self.nodes.items())
        nx_graph.add_nodes_from(nodes_list)
        edges_list: list[tuple[Any, Any, Mapping]] = []
        for edge, data in self.edges.items():
            if self.is_hyperedge(edge):
                us, vs = edge
                if flatten_hyperedges:
                    # flatten the hyperedges into multiple normal edges
                    for u in us:
                        for v in vs:
                            edges_list.append((u, v, {}))
                else:
                    # add a hypernode to connect all in nodes with all out nodes
                    hypernode_id = "_".join(list(map(str, us)) + list(map(str, vs)))
                    nx_graph.add_node(hypernode_id, **data)
                    for u in us:
                        edges_list.append((u, hypernode_id, {}))
                    for v in vs:
                        edges_list.append((hypernode_id, v, {}))
            else:
                u, v = edge  # type: ignore
                edges_list.append((u, v, data))

        nx_graph.add_edges_from(edges_list)
        return nx_graph

    def nodes_of(self, edge: GenericEdge | Nodes | Node) -> Iterator[Node]:
        """Returns an ``Iterator`` of node id's that are incident to the given edge.

        Args:
            edge: an edge of this TrackGraph.

        Yields:
            all nodes incident to the given edge.
        """
        # recursively descent into tuples and yield their elements if they are
        # not tuples
        if isinstance(edge, tuple):
            for x in edge:
                yield from self.nodes_of(x)
        else:
            yield edge

    def is_hyperedge(self, edge: GenericEdge) -> TypeGuard[HyperEdge]:
        """Test if the given edge is a hyperedge in this track graph."""
        assert len(edge) == 2, "(Hyper)edges need to be 2-tuples"
        num_tuples = sum(isinstance(x, tuple) for x in edge)
        if num_tuples == 0:
            return False
        elif num_tuples == 2:
            return True
        raise RuntimeError(
            f"(Hyper)edges have to contain two tuples or two node IDs, not {edge}"
        )

    def _is_hyperedge_nx_node(self, nx_graph: networkx.DiGraph, nx_node: Any) -> bool:
        """Return ``True`` if ``nx_node`` is a hyperedge node in ``nx_graph``.

        Checks if the given networkx node in the given directed networkx graph
        represents an hyperedge. This boils down to checking if the node does not
        have the ``frame_attribute`` set.

        Args:
            nx_graph: a networkx ``DiGraph``.
            nx_node: a node in the given ``nx_graph``.

        Returns:
            bool: true iff the given ``nx_node`` does not posses the
            ``frame_attribute``.
        """
        return self.frame_attribute not in nx_graph.nodes[nx_node]

    def _convert_nx_hypernode(
        self, nx_graph: networkx.DiGraph, hyperedge_node: Any
    ) -> HyperEdge:
        """Creates a hyperedge tuple for hyperedge node in a given networkx ``DiGraph``.

        Args:
            nx_graph: a networkx ``DiGraph``.
            hyperedge_node: a node in the given ``nx_graph`` that represents
                a hyperedge.

        Returns:
            a tuple representing the hyperedge the given ``nx_node`` represented. (It
            will be a tuple with one entry per involved time point, listing all nodes at
            that time point.)
        """
        assert self._is_hyperedge_nx_node(nx_graph, hyperedge_node)

        in_nodes = tuple(nx_graph.predecessors(hyperedge_node))
        out_nodes = tuple(nx_graph.successors(hyperedge_node))

        return (in_nodes, out_nodes)

    def get_frames(self) -> tuple[int, int]:
        """Return tuple with first and last (exclusive) frame this graph has nodes for.

        Returns ``(t_begin, t_end)`` where ``t_end`` is exclusive.
        Returns ``(0, 0)`` for empty graph.
        """
        self._update_metadata()

        return (self.t_begin, self.t_end)

    def nodes_by_frame(self, t: int) -> list[Hashable]:
        """Get all nodes in frame ``t``."""
        self._update_metadata()

        if t not in self._nodes_by_frame:
            return []
        return self._nodes_by_frame[t]

    def _update_metadata(self) -> None:
        if not self._graph_changed:
            return

        self._graph_changed = False
        self._nodes_by_frame: dict[int, list[Hashable]] = {}

        if not self.nodes:
            self._nodes_by_frame = {}
            self.t_begin = 0
            self.t_end = 0
            return

        self._nodes_by_frame = {}
        for node, data in self.nodes.items():
            t = data[self.frame_attribute]
            if t not in self._nodes_by_frame:
                self._nodes_by_frame[t] = []
            self._nodes_by_frame[t].append(node)

        frames = self._nodes_by_frame.keys()
        self.t_begin = min(frames)
        self.t_end = max(frames) + 1

        self._graph_changed = False

    def __str__(self) -> str:
        return f"TrackGraph({len(self.nodes)} nodes, {len(self.edges)} edges)"



from __future__ import annotations

from typing import TYPE_CHECKING

from ..variables import NodeSelected
from .cost import Cost
from .weight import Weight

if TYPE_CHECKING:
    from motile.solver import Solver


class NodeSelection(Cost):
    """Cost for :class:`~motile.variables.NodeSelected` variables.

    Args:
        weight:
            The weight to apply to the cost given by the provided attribute of
            each node. Default is ``1.0``

        attribute:
            The name of the node attribute to use to look up cost, or None if a constant
            cost is desired. Default is ``None``.

        constant:
            A constant cost for each selected node. Default is ``0.0``.
    """

    def __init__(
        self, weight: float = 1, attribute: str | None = None, constant: float = 0.0
    ) -> None:
        self.weight = Weight(weight)
        self.constant = Weight(constant)
        self.attribute = attribute

    def apply(self, solver: Solver) -> None:
        node_variables = solver.get_variables(NodeSelected)

        for node, index in node_variables.items():
            if self.attribute is not None:
                solver.add_variable_cost(
                    index, solver.graph.nodes[node][self.attribute], self.weight
                )
            solver.add_variable_cost(index, 1.0, self.constant)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Collection,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    TypeVar,
)

import ilpy

if TYPE_CHECKING:
    from motile.solver import Solver

_KT = TypeVar("_KT", bound=Hashable)


class Variable(ABC, Mapping[_KT, ilpy.Variable]):
    """Base class for solver variables.

    New variables can be introduced by inheriting from this base class and
    overwriting :func:`instantiate` and optionally
    :func:`instantiate_constraints`.

    Variable classes should not be instantiated by a user. Instead, the
    :class:`~motile.Solver` provides access to concrete variables through the class
    name. The following example shows how to obtain the variable values after
    optimization::

        solver = Solver(graph)

        # add costs, constraints...

        solution = solver.solve()

        # here `node_selected` is an instance of a Variable subclass
        # specifically, it will be an instance of NodeSelected, which
        # maps node Ids to variables in the solver.
        node_selected = solver.get_variables(NodeSelected)

        for node in graph.nodes:
            if solution[node_selected[node]] > 0.5:
                print(f"Node {node} was selected")

    This allows variables to be added lazily to the solver.
    :class:`Constraints<motile.constraints.Constraint>` and
    :class:`motile.costs.Cost` can ask for variables.
    """

    # default variable type, replace in subclasses to override
    variable_type: ClassVar[ilpy.VariableType] = ilpy.VariableType.Binary

    @staticmethod
    @abstractmethod
    def instantiate(solver: Solver) -> Collection[_KT]:
        """Create and return keys for the variables.

        For example, to create a variable for each node, this function would
        return a list of all nodes::

            @staticmethod
            def instantiate(solver):
                return solver.graph.nodes

        The solver will create one variable for each key. The index of that
        variable can be accessed through a dictionary returned by
        :meth:`motile.Solver.get_variables`::

            solver = Solver(graph)

            node_selected = solver.get_variables(NodeSelected)

            for node in graph.nodes:
                index = node_selected[node]
                print(f"Selection indicator of node {node} has index {index}")

        Args:
            solver:
                The :class:`~motile.Solver` instance to create variables for.

        Returns:
            A collection of keys (anything that is hashable, e.g., nodes of a graph),
            one for each variable to create.
        """
        pass

    @staticmethod
    def instantiate_constraints(
        solver: Solver,
    ) -> Iterable[ilpy.Constraint | ilpy.Expression]:
        """Add constraints for this variable to the solver.

        This ensures that these variables are coupled to other variables of the solver.

        Args:
            solver:
                The :class:`~motile.Solver` instance to create variable constraints for.

        Returns:
            A iterable of :class:`ilpy.Constraint` or
            :class:`ilpy.expressions.Expression.` See
            :class:`motile.constraints.Constraint` for how to create linear constraints.
        """
        return []

    def __init__(self, solver: Solver, index_map: dict[_KT, int]) -> None:
        self._solver = solver
        self._index_map = index_map

    def __repr__(self) -> str:
        rs = []
        for key, index in self._index_map.items():
            r = f"{type(self).__name__}({key}): "
            if index < len(self._solver.costs):
                r += f"cost={self._solver.costs[index]} "
            else:
                r += "cost=None "
            if self._solver.solution is not None and index < len(
                self._solver.solution.variable_values
            ):
                r += f"value={self._solver.solution[index]}"
            else:
                r += "value=None"
            rs.append(r)
        return "\n".join(rs)

    def __getitem__(self, key: _KT) -> ilpy.Variable:
        name = f"{type(self).__name__}({key})"
        return ilpy.Variable(name, index=self._index_map[key])

    def __iter__(self) -> Iterator[_KT]:
        return iter(self._index_map)

    def __len__(self) -> int:
        return len(self._index_map)

    # All of these methods are provided by subclassing typing.Mapping
    # __contains__
    # keys
    # items
    # values
    # get
    # __eq__
    # __ne__



from __future__ import annotations

from typing import Any, Mapping, TypeAlias, Union

# Nodes are integers
Node: TypeAlias = int
# Collections of nodes (for hyperedges) are tuples
Nodes: TypeAlias = tuple[int, ...]

# Edges are tuples of Node or Nodes.
# (0, 1) is an edge from node 0 to node 1.
# ((0,), (1, 2)) is a hyperedge from node 0 to nodes 1 and 2 (i.e. a split).
# ((0, 1), 2) is not valid.
Edge: TypeAlias = tuple[Node, Node]
HyperEdge: TypeAlias = tuple[Nodes, Nodes]
GenericEdge: TypeAlias = Union[Edge, HyperEdge]

# objects in the graph are represented as dicts
# eg. { "id": 1, "x": 0.5, "y": 0.5, "t": 0 }
Attributes: TypeAlias = Mapping[str, Any]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from motile.solver import Solver


class Cost(ABC):
    """A base class for a cost that can be added to a solver.

    Weights should be initialized in the __init__ and added to a instance
    variable so that the Solver can discover them.
    """

    @abstractmethod
    def apply(self, solver: Solver) -> None:
        """Apply a cost to the given solver.

        Use :func:`motile.Solver.get_variables` and
        :func:`motile.Solver.add_variable_cost`.

        Args:
            solver:
                The :class:`~motile.Solver` to create a cost for.
        """
        pass
